import random
import Function.Function as FunctionModule
import json
import os

class Simulation:
    
    # Input: InputDomainSize, MaximumInverseImageSize, which are for Simulation_asDescribedInPaper
    # Process: This method first runs Simulation_asDescribedInPaper to generate a set of random inverse image sizes.
    #          These inverse image sizes can be abstracted into a function and probability mass function that assumes a
    #          uniform distribution; the method therefore does this. The method then computes Sq, DRR, PColl, and NSq on
    #          this function and probability mass function and stores the results in a SimulationData object.
    # Output: This data object is finally returned.
    # Note: Additions were made for the experiment manager - in particular ExperimentFolder and StoreData determines whether we store the simulation data and 
    # where, and Experiment_Type is used to change how we conduct the experiment. ConstantInputDomainSize is what was originally used in the IPL 2012 paper...
    # in this case, the simulation will have the specified size. If RandomInputDomainSize is chosen, then the specified input domain size is merely an upper bound
    def runSimulation(self, InputDomainSize, MaximumInverseImageSize,ExperimentFolder="N/A",StoreData=False,Experiment_Type="ConstantInputDomainSize"):
        
        InverseImageSizes = None 
        if Experiment_Type == "ConstantInputDomainSize": # Used in IPL 2012 replication
            InverseImageSizes = self.Simulation_asDescribedInPaper(InputDomainSize, MaximumInverseImageSize)
        elif Experiment_Type == "RandomInputDomainSize":
            InverseImageSizes = self.Simulation_asDescribedInPaper(random.randint(1,InputDomainSize), MaximumInverseImageSize)
        elif Experiment_Type == "RandomInputDomainSize_RandomMaximumInverseImageSize":
            InverseImageSizes = self.Simulation_asDescribedInPaper(random.randint(1,InputDomainSize), random.randint(1,MaximumInverseImageSize))
        
        Function = self.generateInputToOutputMappings(InverseImageSizes)
        ProbabilityMassFunction = self.generateUniformProbabilityMassFunction(list(Function.keys()))
        FunctionObject = FunctionModule.Function(Function,ProbabilityMassFunction)
        
        SimuData = SimulationData(Function,
                              ProbabilityMassFunction,
                              FunctionObject.compute_Squeeziness_MethodOne(),
                              FunctionObject.compute_Domain_to_Range_Ratio(),
                              FunctionObject.compute_Probability_of_Collision(),
                              FunctionObject.compute_Normalised_Squeeziness()
                        )
        
        if StoreData == True:
            SimuData.saveSimulationData_JSON(ExperimentFolder)
        
        return SimuData
         
    
    # Input: InputDomainSize is the desired size of the input domain
    #        MaximumInverseImageSize is an upper bound on the number of inputs that can map on to a single output
    # Process: Given an InputDomainSize, this method partitions InputDomainSize into N random partitions, such that
    #          the maximum size of a partition can bet MaximumInverseImageSize, and the minimum size is 1
    # Output: InverseImageSizes... which is a list of image sizes for a random variable.
    def Simulation_asDescribedInPaper(self, InputDomainSize, MaximumInverseImageSize):
        InverseImageSizes = []    
        while sum(InverseImageSizes) != InputDomainSize:
            InverseImageSize = random.randint(1,MaximumInverseImageSize)
            
            if (sum(InverseImageSizes)+InverseImageSize) <= InputDomainSize:
                InverseImageSizes.append(InverseImageSize)                
            else:
                InverseImageSizes.append(InputDomainSize-sum(InverseImageSizes))
        return InverseImageSizes

    # Input: InverseImageSizes, which is a list of integers
    # Process: Uses the InverseImageSizes to create a dictionary that maps inputs to outputs, such that these inputs
    #          and outputs are compliant with the InverseImageSizes
    # Output: A dictionary that maps inputs to outputs... or in other words, the input and output domains.
    def generateInputToOutputMappings(self,InverseImageSizes):
        Function = {}
        CurrentInput = 1
        CurrentOutput = 1
        for InverseImage in InverseImageSizes:
            for _ in range(InverseImage):
                Function[CurrentInput] = CurrentOutput
                CurrentInput = CurrentInput + 1
            CurrentOutput = CurrentOutput + 1
        return Function

    # Input: a list that pertains to the input domain of a function
    # Process: computes a probability mass function for the input domain, based on the uniform distribution
    # Output: returns a dictionary that pertains to the probability mass function
    def generateUniformProbabilityMassFunction(self, Inputs):
        Probability = 1/len(Inputs)
        ProbabilityMassFunction = {}
        for Input in Inputs:
            ProbabilityMassFunction[Input] = Probability
        return ProbabilityMassFunction


# This is a data object for a single simulation's data.
class SimulationData():
    
    def __init__(self,Function,ProbabilityMassFunction,Squeeziness,DomainToRangeRatio,ProbabilityOfCollisions,NormalisedSqueeziness):
        self.Function = Function
        self.ProbabilityMassFunction = ProbabilityMassFunction
        self.Squeeziness = Squeeziness
        self.DomainToRangeRatio = DomainToRangeRatio
        self.ProbabilityOfCollisions = ProbabilityOfCollisions
        self.NormalisedSqueeziness = NormalisedSqueeziness
    
    # This stores the simulation data in JSON format at the specified folder
    def saveSimulationData_JSON(self,ExperimentFolder):
        
        JSONData = {"Funct":self.Function,
                    "PMF":self.ProbabilityMassFunction,
                    "Sq":self.Squeeziness,
                    "DRR":self.DomainToRangeRatio,
                    "PColl":self.ProbabilityOfCollisions,
                    "NSq":self.NormalisedSqueeziness
                    }

        ParentFolder = ExperimentFolder+"\\Simulations"
        if os.path.exists(ParentFolder) == False:
            os.mkdir(ParentFolder)
        JSONFileLocation = ParentFolder + "\\Sim_" + str(len(os.listdir(ParentFolder))+1)+".json"
        if os.path.exists(JSONFileLocation):
            input("This file already exists: "+JSONFileLocation+". Are you sure you want to continue?")
        
        JSONFileObject = None
        try:
            JSONFileObject = open(JSONFileLocation,"w+")
            json.dump(JSONData,JSONFileObject,indent=4)
        except:
            raise Exception("Unable to write simulation data.")
        finally:
            if JSONFileObject != None:
                JSONFileObject.close()

            
        












